#|
Arxiu: estadistica.lisp.
Contingut: mètodes d'estadística.
Autor: Felip Alàez Nadal.
Útima actualització 3-VI-06

|#

#|BSD License

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. |#
       



#| Mòdul d'estadística. Conté funcions que calculen magnituts estadítiques, com la mitja, desviació estàdar, desviació estàndar muestra, etc... d'una mostra de dades. |#

(provide :ph_estatistical)

(defpackage :ph_estatistical
  (:use "CL")
  (:nicknames "ESTADISTICA")
  
  (:export #:MITJA #:DESVIACIO_ESTANDAR #:DESVIACIO_ESTANDAR_MOSTRAL #:DISPERSIO #:DISPERSIO_PERCENTUAL)
  (:documentation "This package provides statistical functions.")
  )

(in-package ph_estatistical)

(defun mitja (dades)
  "Computes the mean of the numbers at a list."
  (let ((suma 0))
    (setq suma (apply #'+ dades))
    (/ suma (length dades))
    )
  )


(defun desviacio_estandar (dades)
  "Computes the standard deviation of dades ( a list)"
  (let ((mitja 0)(sum 0))
    (setq mitja (mitja dades))
    (setq sum (apply #'+ (mapcar #'(lambda (x) (expt (- x mitja) 2)) dades)))
    (/ (expt sum  1/2) (expt (length dades) 1/2))
    )
  )


(defun desviacio_estandar_mostral (dades)
  "Computes the mostral standard deviation of dades"
  (let ((aux 0))
    (setq aux (* (desviacio_estandar dades) (expt (length dades) 1/2)))
    (setq aux (/ aux (expt (- (length dades) 1) 1/2)))
  )
)


(defun dispersio (dades)
  "Returns the dispersion of dades"
  (let ((max 0) (min 0))
    (setq max (apply #'max dades))
    (setq min (apply #'min dades))
    (- max min)
    )
  )


(defun dispersio_percentual (dades)
  "Returns the percent dispersion of dades."
  (* 100 (/ (dispersio dades) (mitja dades)))
  )
  
