#|
Arxiu: ph_integration.lisp
Contingut: implements some integration methods.
Autor: Felip Alàez Nadal.
Útima actualització 25-VII-06

|#

#|BSD License

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. |#

(defpackage :ph_integration 
  (:use :cl)
  (:export #:trapezoidal)
  (:documentation "This package provides methods for integration, other than montecarlo's"))

(in-package :ph_integration)

(defun trapezoidal ( f a b &optional ( n_intervals 100 ) )
  "Computes the integral of f between a and b using n_points and the multiple trapezoidal rule."
  (unless (> b a) (return-from trapezoidal 'error))
  (let
      ( ( h (/ (abs (- b a))  n_intervals  ))
       ( tmp a)
	( rst 0)
	)
    (loop
     (if (<= b tmp) (return))
     (incf tmp h)
     (incf rst (* (funcall f tmp) h))
     )
    (coerce rst 'float )
    )
  )