#|
Arxiu: ph_pseudoaleatorial.lisp
Contingut: funcions per a la generació de nombres pseudoaleatoris.
Autor: Felip Alàez Nadal.
Útima actualització: 25-II-07

|#

#|BSD License

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. |#

(declaim (optimize (speed 3) (debug 0) (safety 0) (compilation-speed  0) (space 0)))

(defpackage :ph_random
  (:use :cl )
  (:export #:carlson_init #:carlson)
  (:documentation "Provides several pseudoaleatorial number generators."))


(in-package :ph_random)


(let* (
      (vmodul_ini 83886080.)
      (vmodul  (* 2. vmodul_ini) )
      (y_k 54)
      (y_ant 23)
      (y (make-array 55)))
      
  (defun carlson_init (seed)
    "Initializes the Carlson pseudorandom numbers generator."
    (setf (aref y 0) seed)
    (do ((i 1 (incf i)))
	((= i 55) (values))
      (setq seed (mod (* seed 16805.) vmodul_ini))
      (setf (aref y i) (coerce (+ seed (mod (* 16805. (aref y (- i 1))) vmodul_ini)) 'double-float)))
    )


  (defun carlson ()
    "Returns a random number from the uniform U(0,1) pdf using a modification of the Carlson Algorithm."
    (let
	(
	 (retornable (mod (+ (aref y y_k) (aref y y_ant)) vmodul))
	 )
      (setf (aref y y_k) retornable)
      (decf y_k)
      (decf y_ant)
      (cond
	((= y_k -1) (setq y_k 54))
	((= y_ant -1) (setq y_ant 54))
	)
     (coerce  (/  retornable vmodul) 'double-float)
      )
    ))


