#|BSD License

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. |#



(require :util )

(defpackage :ph_simplex
  (:use :cl :util)
  (:export #:crea_simplex #:min_simplex)
  (:documentation "This package provides the simplex n-variable minimization method.")
)

(in-package :ph_simplex)

(defun ordena_simplex (simplex funcio)
  "Sorts the simplex. Its first element will now be the more little, and the last the bigger.Its paramethers are the simplex and a function used to sort, which returns < 0 if its first argument is more little than the second, and > 0 otherwise."
  (let ((simp (copy-seq simplex))) ;copie la seqüència per a no cagar-la, ggg
	(reverse (sort simp (lambda (x y) (if (< (apply funcio x) (apply funcio y)) nil t))))))

(defun centre_de_gravetat (simp )
  " Returns the gravity center of a simplex. The simplex must be sorted."
  (let* ((primer (nth 0 simp))
	 (dimensio (length primer))
					;(elements (+ dimensio 1)) ; en dimensio 1 el símplex té 2 elements, en dimensió 2, 3 , etc... Per raons d'optimització use dimensio enlloc i no necessite aquesta variable.
	 (i 0)
	 (cdg (repeat 0 dimensio)) ;inicialitze el centre de gravetat al vector nul en el nombre de dimensions que usem.
	 )
    (loop
     (if (eq i dimensio) (return) ;L'ltim vector que sumarem és el dimensio - 1 (el que ocupa la posició dimensio és el pitjor)
       (progn 
	 (setq cdg (mapcar #'+ cdg (nth i simp)))
	 (incf i)
	 )))
    (mapcar #'* (repeat (/ 1 i) dimensio ) cdg) ; Ara multiplique cdg per 1/i , per obtindre el resultat final.

     ))



(defun crea_simplex (n_dimensions &rest valors_inicials)
  "Returns a n-dimension simplex, which uses valors_inicials."
  (let ((simplex '())
	aux)
    (do ((i 0 (incf i)))
	((= i  n_dimensions ) t)
	(setq aux (repeat 0 n_dimensions))
	(setf (nth i aux) 1)
	(push aux simplex))
    (push valors_inicials simplex)))
    

(defun min_simplex (simplex funcio &key (NITER 100) (tol_rel 1e-3))
  "Minimizes the function using simplex. You can change the maximum number of iterations and the relative tolerance."
  (let* ((smp (copy-seq simplex)) ;Copie el símplex per a evitar modificar-lo. 
	(err 0) ;Variable on posaré l'error màxim (en valor absolut) que es tolleraré
	(iteracio 0) ;Variable per guardar el nombre d'iteració actual.
	(dimensio (length (nth 0 smp)))
	(pos_millor 0)
	(pos_pitjor dimensio)
	(pos_quasipitjor (- dimensio 1))
	millor
	pitjor
	quasipitjor
	cdg
	nou
	nou2
	)

    (setq smp (ordena_simplex smp funcio))
    (setq millor (nth pos_millor smp))
    (setq pitjor (nth pos_pitjor smp))
    ;Assigne l'error com , amb els valors inicials de millor i pitjor, (f(millor) - f(pitjor))*tol_rel
    (setq err (abs (* (- (apply funcio millor) (apply funcio pitjor)) tol_rel)))
;    (format t "Treballant amb un error ~A~%" err)
    (loop
     (if (= iteracio NITER) (return) 
       (progn
	
	 (setq smp (ordena_simplex smp funcio))
	 (setq cdg (centre_de_gravetat smp))
	 (setq millor (nth pos_millor smp))
	 (setq pitjor (nth pos_pitjor smp))
	 (setq quasipitjor (nth pos_quasipitjor smp))
	 (incf iteracio)
					;Una vegada calculades totes les magnituts necessàries, em pose a calcular.
					;El punt nou es calcula com :nou = 2*cdg - pitjor;
	 (setq nou (mapcar #'- (mapcar #'* (repeat 2 dimensio) cdg) pitjor))
					;Cas 1:
	 (if (< (apply funcio nou) (apply funcio millor))
	     (progn ;En aquest cas es calcula una altra reflexió	       (setq nou2 (mapcar #'- (mapcar #'* (repeat 3 dimensio) cdg) pitjor))
;	       (format t "Cas 1~%")
	       (if (< (apply funcio nou2) (apply funcio nou)) ;Cas 1A: En aquest cas tenim un punt molt bo i el canviem el pitjor per nou2
		   (progn
		 ;    (format t "Cas 1A~%")
		     (setf (nth pos_pitjor smp) nou2))
		 (progn ;Cas 1B: En cas contrari posem nou a la posició de pitjor
		;   (format t "Cas 1B~%")
		   (setf (nth pos_pitjor smp) nou))))
	   (progn 
	     (if (< (apply funcio nou) (apply funcio quasipitjor)) ;Cas 2: el nou punt és millor que quasipitjor -> canviem nou per pitjor
		 (progn 
	;	   (format t "Cas 2~%")
		   (setf (nth pos_pitjor smp) nou)))
	     (progn ;Cas 3: El nou punt és pitjor que quasipitjor
;	       (format t "Cas 3~%")
					;Calcule nou2 com nou2 = 0.5*(cdg + pitjor):
	       (setq nou2 (mapcar #'* (repeat 0.5 dimensio) (mapcar #'+ cdg pitjor)))
					;Cas 3a: el nou punt � millor que quasipitjor -> canviem pitjor per nou2
	       (if (< (apply funcio nou2) (apply funcio quasipitjor))
		   (progn 
		 ;    (format t "Cas 3A~%")
		     (setf (nth pos_pitjor smp) nou2))
		 (progn ;Cas 3b: procedim a una contracció general del simplex
		;   (format t "Cas 3B~%")
		   (do ((contador 1 (incf contador)))
		       ((= contador (+ dimensio 1)) t) 
		       (setf (nth contador smp) (mapcar #'* (repeat 0.5 dimensio) (mapcar #'+ millor (nth contador smp))))))))))
	 (setq smp (ordena_simplex smp funcio))
	 (let ((error_pas (abs (- (apply funcio millor) (apply funcio pitjor)))) ;error_pas és l'error que cometem a aquest pas
	       (error_permes (max tol_rel (* tol_rel (/ (+ (apply funcio millor) (apply funcio pitjor)) 2)))))
;	   (format t "L'error en aquest pas és ~A i l'error admissible ~A~%" error_pas error_permes)
	   (if (< (- error_pas error_permes) 0) ;Pense que és millor el nou mèode per calcular l'error, i amés que cal fer la resta, doncs comparar dos nombres molt semblants pot donar problemes
	       (progn
;		 (print "He trobat un mínim")
		 (return-from min_simplex (values t smp (mapcar #'(lambda (x) (if (= 0 x) error_pas (abs x))) (mapcar #'- (nth pos_millor smp) (nth pos_pitjor smp)))))))
	   )
	 )))
;    (print smp)
    (values nil smp)))


